##---------------------------------------------------------------------------##
##  File:
##      readmail.pl
##  Author:
##      Earl Hood       ehood@convex.com
##  Description:
##      MHonArc library defining routines to parse e-mail messages.
##---------------------------------------------------------------------------##
##    MHonArc -- Internet mail-to-HTML converter
##    Copyright (C) 1995	Earl Hood, ehood@convex.com
##
##    This program is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program; if not, write to the Free Software
##    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##---------------------------------------------------------------------------##

package readmail;

##---------------------------------------------------------------------------##
##  Notes:
##	base64.pl and qprint.pl must be required before-hand inorder for
##	MAILread_body() to work.
##---------------------------------------------------------------------------##

##---------------------------------------------------------------------------
##	MAILread_file_header reads (and strips) a mail message header from the
##	filehandle $handle.  *mesg is a pointer to the mail message.  *fields
##	is a pointer to an associative array to put field values in indexed
##	by field labels that have been converted to all lowercase.  *l2o
##	is an associative array to get the original label text from the
##	lowercase field label keys.  The return value is the original
##	header text.
##	
sub main'MAILread_file_header {
    local($handle, *fields, *l2o) = @_;
    local($label, $olabel, $value, $tmp, $header);
    local($d) = ($/);

    $/ = "\n";  $label = '';
    $header = '';  %fields = ();  %l2o = ();
    while (($tmp = <$handle>) !~ /^[\r]?$/) {
	$header .= $tmp;
	$tmp =~ s/\n/ /g;

	## Check for continuation of a field
	if ($tmp =~ /^\s/) {
	    $fields{$label} .= $tmp  if $label;
	    next;
	}

	## Separate head from field text
	($olabel, $value) = $tmp =~ /^([^:]+):\s*([\s\S]*)$/;
	($label = $olabel) =~ tr/A-Z/a-z/;
	$l2o{$label} = $olabel;
	if ($fields{$label}) {
	    $fields{$label} .= $'X . $value;
	} else {
	    $fields{$label} = $value;
	}
    }
    $/ = $d;
    $header;
}
##---------------------------------------------------------------------------
##	MAILread_body() parses a MIME message body.  $header is the
##	header of the message.  $content is the value of the Content-Type
##	field.  $body is the actual message body.  The return value
##	is an array:  The first item is the HTML generated, and
##	the other items are filenames of any derived files.
##
sub main'MAILread_body {
    local($header, *fields, $body) = @_;
    local($type, $subtype, $boundary, $part, $parthead,
	  $ret, $tmp, $content, $ctype);
    local(@parts);
    local(%partfields);
    local(%partl2o);
    local(@files);
    local(@array);

    ## Get type/subtype
    $content = $fields{'content-type'} || 'text/plain';
    ($ctype) = $content =~ m%^\s*([\w-/]+)%;
    $ctype =~ tr/A-Z/a-z/;
    if ($ctype =~ m%/%) {
	($type,$subtype) = split(/\//, $ctype, 2);
    } elsif ($ctype =~ /text/) {
	$ctype = 'text/plain';
	$type = 'text';  $subtype = 'plain';
    } else {
	$type = $subtype = '';
    }

    ## If multipart, recursively process each part
    if ($type =~ /multipart/i) {

	## Get boundary
	if ($content =~ m%boundary\s*=\s*"([^"]*)"%i) {
	    $boundary = $1;
	} else {
	    ($boundary) = $content =~ m%boundary\s*=\s*(\S+)%;
	}
	$boundary =~ s/(\W)/\\$1/g;

	## Split parts and process each
	$body = "\r\n" . $body;
	if ($subtype =~ /alternative/i) {	# go in reverse order
	    @parts = reverse split(/\r?\n--$boundary/, $body);
	    pop @parts;
	    while (@parts && ($parts[0] !~ /^--/)) { shift @parts; }
	    shift @parts;
	} else {
	    @parts = split(/\r?\n--$boundary/, $body);
	    shift @parts;
	    while (@parts && ($parts[$#parts] !~ /^--/)) { pop @parts; }
	    pop @parts;
	}

	## Process parts
	foreach $part (@parts) {
	    $part =~ s/^\r?\n//;	# Drop begining newline

	    ## Read header
	    $parthead = &'MAILread_header(*part, *partfields, *partl2o);

	    ## Get content type
	    $partcontent = $partfields{'content-type'};
	    if (!$partfields{'content-type'}) {
		if ($subtype =~ /digest/) {
		    $partfields{'content-type'} = 'message/rfc822'; }
		else { $partfields{'content-type'} = 'text/plain'; }
	    }

	    ## Only use last filterable part in alternate
	    if ($subtype =~ /alternative/) {
		@array = &'MAILread_body($parthead, *partfields, $part);
		$ret = shift @array;
		if ($ret) {
		    push(@files, @array);
		    last;
		}
	    } else {
		@array = &'MAILread_body($parthead, *partfields, $part);
		if (!$array[0]) {
		    $array[0] = join('',
				"<P><UL><LI>\n",
				"Could not process part with given ",
				"Content-Type: \n",
				"<code>", $partfields{'content-type'},
				"</code>\n",
				"</UL>\n"
				);
		}
		$ret .= shift @array;
	    }
	    push(@files, @array);
	}
	if (!$ret && ($subtype =~ /alternative/)) {
	    $ret = join('',
		    "<hr>\n",
		    "No recognizable part in ",
		    "<code>multipart/alternative</code> part.\n",
		    "<hr>\n");
	}

    ## Message/rfc822 or Message/news
    } elsif ($ctype =~ m%message/(rfc822|news)%) {
	$parthead = &'MAILread_header(*body, *partfields, *partl2o);
	$ret = join('',
		    "<em>\n",
		    "-- BEGIN included message.\n",
		    "</em><p>\n",
		    "<BLOCKQUOTE>\n");
	$ret .= &'htmlize_header(*partfields, *partl2o);

	@array = &'MAILread_body($parthead, *partfields, $body);
	$ret .= shift @array ||
		join('',
		     "<P><UL><LI>\n",
		     "Could not process part with given Content-Type: \n",
		     "<code>", $partfields{'content-type'}, "</code>\n",
		     "</UL>\n");
	$ret .= join('',
		    "</BLOCKQUOTE>\n",
		    "<em>\n",
		    "-- END included message\n",
		    "</em><p>\n");
	push(@files, @array);

    ## Non-multipart, call registered filter of specified content-type.
    } else {
	## Call filter
	$filter = $'MIMEFilters{"$ctype"};	# MIMEFilters defined by main
	if ($filter) {
	    ## Check encoding
	    local($encoding) = $fields{'content-transfer-encoding'};
	    local($decoded);
	    if ($encoding =~ /base64/i) {
		$decoded = &base64'b64decode($body);
		@array = &$filter($header, *fields, *decoded);
	    } elsif ($encoding =~ /quoted-printable/i) {
		$decoded = &quoted_printable'qprdecode($body);
		@array = &$filter($header, *fields, *decoded);
	    } else {
		@array = &$filter($header, *fields, *body);
	    }
	    $ret = shift @array;
	} else {
	    $ret = '';	# Return null string if no filter to process.
	}
	push(@files, @array);
    }
    ($ret, @files);
}
##---------------------------------------------------------------------------
##	MAILread_header reads (and strips) a mail message header from the
##	variable *mesg.  *mesg is a pointer to the mail message.  *fields
##	is a pointer to an associative array to put field values in indexed
##	by field labels that have been converted to all lowercase.  *l2o
##	is an associative array to get the original label text from the
##	lowercase field label keys.  The return value is the original
##	header text.
##	
sub main'MAILread_header {
    local(*mesg, *fields, *l2o) = @_;
    local($label, $olabel, $value, $tmp, $header);

    $header = '';  %fields = ();  %l2o = ();  $label = '';

    ## Read a line at a time.
    while ($mesg =~ s/^([^\n]*\n)//) {
	$tmp = $1;
	last  if $tmp =~ /^[\r]?$/;	# Done if blank line
	$header .= $tmp;
	$tmp =~ s/\n/ /g;

	## Check for continuation of a field
	if ($tmp =~ /^\s/) {
	    $fields{$label} .= $tmp  if $label;
	    next;
	}

	## Separate head from field text
	($olabel, $value) = $tmp =~ /^([^:]+):\s*([\s\S]*)$/;
	($label = $olabel) =~ tr/A-Z/a-z/;
	$l2o{$label} = $olabel;
	if ($fields{$label}) {
	    $fields{$label} .= $'X . $value;
	} else {
	    $fields{$label} = $value;
	}
    }
    $header;
}

##---------------------------------------------------------------------------##
1;
